"""Module for random functions in jumpy."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as onp

from jumpy.core import which_np

if TYPE_CHECKING:
    from jumpy.numpy import ndarray

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax, jnp = None, None


__all__ = [
    "PRNGKey",
    "uniform",
    "split",
    "randint",
    "choice",
]


def PRNGKey(seed: int) -> ndarray:
    """Returns a PRNG key given a seed."""
    # NOTE: selects backend based on seed type.
    if which_np(seed) is jnp:
        return jax.random.PRNGKey(seed)
    else:
        rng = onp.random.default_rng(seed)
        return rng.integers(low=0, high=2**32, dtype="uint32", size=2)


def uniform(
    rng: ndarray,
    shape: tuple[int, ...] = (),
    low: float | None = 0.0,
    high: float | None = 1.0,
) -> ndarray:
    """Sample uniform random values in [low, high) with given shape/dtype."""
    if which_np(rng) is jnp:
        return jax.random.uniform(rng, shape=shape, minval=low, maxval=high)
    else:
        return onp.random.default_rng(rng).uniform(size=shape, low=low, high=high)


def split(rng: ndarray, num: int = 2) -> ndarray:
    """Splits a PRNG key into num new keys by adding a leading axis."""
    if which_np(rng) is jnp:
        return jax.random.split(rng, num=num)
    else:
        rng = onp.random.default_rng(rng)
        return rng.integers(low=0, high=2**32, dtype="uint32", size=(num, 2))


def randint(
    rng: ndarray,
    shape: tuple[int, ...] = (),
    low: int | None = 0,
    high: int | None = 1,
) -> ndarray:
    """Sample integers in [low, high) with given shape."""
    if which_np(rng) is jnp:
        return jax.random.randint(rng, shape=shape, minval=low, maxval=high)
    else:
        return onp.random.default_rng(rng).integers(low=low, high=high, size=shape)


def choice(
    rng: ndarray,
    a: int | ndarray,
    shape: tuple[int, ...] = (),
    replace: bool = True,
    p: Any | None = None,
    axis: int = 0,
) -> ndarray:
    """Generate sample(s) from given array."""
    if which_np(rng) is jnp:
        return jax.random.choice(rng, a, shape=shape, replace=replace, p=p, axis=axis)
    else:
        return onp.random.default_rng(rng).choice(
            a, size=shape, replace=replace, p=p, axis=axis
        )



"""Jumpy module."""

try:
    import jax.numpy as jnp  # noqa: F401

    is_jax_installed = True
except ImportError:
    is_jax_installed = False


from jumpy import core, lax, ops, random
from jumpy._base_fns import index_update, vmap

__all__ = [
    # === primitives ===
    # === sub-modules ===
    "core",
    "random",
    "numpy",
    "lax",
    "ops",
    # === base functions ==
    "index_update",
    "vmap",
]



"""Module for linear algebra for jumpy."""

from __future__ import annotations

import numpy as onp

from jumpy.core import which_np
from jumpy.numpy._types import ndarray

try:
    import jax.numpy as jnp
except ImportError:
    jnp = None


__all__ = ["inv", "norm", "safe_norm"]


def inv(a: ndarray) -> ndarray:
    """Compute the (multiplicative) inverse of a matrix."""
    return which_np(a).linalg.inv(a)


def norm(x: ndarray, axis: tuple[int, ...] | int | None = None) -> ndarray:
    """Returns the array norm."""
    return which_np(x, axis).linalg.norm(x, axis=axis)


def safe_norm(x: ndarray, axis: tuple[int, ...] | int | None = None) -> ndarray:
    """Calculates a linalg.norm(x) that's safe for gradients at x=0.

    Avoids a poorly defined gradient for jnp.linalg.norm(0) see
    https://github.com/google/jax/issues/3058 for details

    Args:
      x: A jp.ndarray
      axis: The axis along which to compute the norm

    Returns:
      Norm of the array x.
    """
    np = which_np(x)
    if np is jnp:
        is_zero = jnp.allclose(x, 0.0)
        # temporarily swap x with ones if is_zero, then swap back
        x = jnp.where(is_zero, jnp.ones_like(x), x)
        n = jnp.linalg.norm(x, axis=axis)
        n = jnp.where(is_zero, 0.0, n)
    else:
        n = onp.linalg.norm(x, axis=axis)
    return n



"""Core functions for jumpy."""

from __future__ import annotations

import builtins
from typing import Any

import numpy as onp

import jumpy as jp

__all__ = [
    "is_jitted",
    "which_np",
    "which_dtype",
    "custom_jvp",
]

try:
    import jax
    import jax.numpy as jnp
    from jax import custom_jvp, tree_map
    from jax.interpreters.batching import BatchTracer
except ImportError:
    jax, jnp = None, None

    BatchTracer = None
    tree_map = None

    class custom_jvp:
        """Custom JVP decorator."""

        def __init__(self, func: callable):
            """Initialise the custom jvp with func."""
            self.func = func

        def __call__(self, *args, **kwargs):
            """Calls the func with `args` and `kwargs`."""
            return self.func(*args, **kwargs)

        def defjvp(self, func):
            """For numpy only, we ignore the defjvp function."""


def is_jitted() -> bool:
    """Returns true if currently inside a jax.jit call or jit is disabled."""
    if jp.is_jax_installed is False:
        return False
    elif jax.config.jax_disable_jit:
        return True
    else:
        return jax.core.cur_sublevel().level > 0


def which_np(*args: Any):
    """Returns which numpy implementation (Numpy or Jax) based on the arguments."""
    if jp.is_jax_installed is False:
        return onp

    checker = lambda a: (  # noqa: E731
        isinstance(a, (jnp.ndarray, BatchTracer)) and not isinstance(a, onp.ndarray)
    )
    if builtins.any(jax.tree_util.tree_leaves(tree_map(checker, args))):
        return jnp
    else:
        return onp


def which_dtype(dtype: object | None):
    """Returns np or jnp depending on dtype."""
    if jp.is_jax_installed and dtype is not None and dtype.__module__ == "jax.numpy":
        return jnp
    else:
        return onp
